"""
This module is used by solvers and models mostly
"""
from typing import Tuple, List, Sequence, Hashable
from ._typevars import Task, Result
from .cache import BaseCache


def tasks_sort(tasks: Sequence[Task], cache: BaseCache[Task, Result]) -> Tuple[List[int], List[int], List[int]]:
    """Sorts the tasks list into:
        * those that need to be solved
        * those that are already in the cache
        * those that are repeated in the tasks list
    and returns three lists of task indexes

    :param tasks: List of tasks
    :param cache: BaseCache
    :return: tasks indexes sorted into three lists
    :raises ValueError: raises if any task is not Hashable
    """
    to_solve, cached, same = [], [], []
    to_solve_hash = []
    for i, task in enumerate(tasks):
        if not isinstance(task, Hashable):
            raise ValueError('If caching is True all the tasks must be hashable')
        if task in cache:
            cached.append(i)
        elif task in to_solve_hash:
            same.append(i)
        else:
            to_solve.append(i)
            to_solve_hash.append(task)
    return to_solve, cached, same



"""
This module contains Task class
"""
from typing import Hashable
from warnings import warn


class Task:
    """Task class"""
    def __init__(self, *args, tag=None, **kwargs):
        warn("Task class will be removed soon. Please use custom tasks.", DeprecationWarning, stacklevel=2)
        if not isinstance(tag, Hashable):
            raise TypeError('tag must be Hashable')
        self.tag = tag
        self.args = args
        self.kwargs = kwargs

    def __hash__(self):
        return hash(self.tag)

    def __eq__(self, other):
        return hash(self.tag) == hash(other.tag)



"""
A simple package for parallel computing with Python
"""
from .models import (
    Model
)
from .workers import (
    Worker,
    MultiprocessingWorker,
    MPIWorker,
    MPIWorker
)
from .solvers import (
    MultiprocessingSolver,
    SimpleSolver,
    MPISolver,
    Solver,
    get_solver
)

from .task import (
    Task
)

from .cache import (
    BaseCache,
    DictCache
)



"""
This module contains different Workers
"""
import multiprocessing
from multiprocessing import JoinableQueue, Queue
import socket
import logging
from typing import Generic

from .models import Model
from .task import Task as OldTask
from ._typevars import Task, Result

logger = logging.getLogger(__package__ + '.worker')


class Worker(Generic[Task, Result]):
    """
    The Worker abstract base class.
    """
    def __init__(
            self,
            model: Model[Task, Result]
    ):
        self.model = model

    def start(self):
        """
        Function that stars the worker.
        Always called on each worker by solver when initializing.

        :return:
        """

    def do_the_job(self, task: Task) -> Result:
        """
        calls Model.results() with specified params

        :param task: task for Model.results()
        :return: pass Model.results()
        """
        if isinstance(task, OldTask):
            return self.model.results(*task.args, **task.kwargs)
        return self.model.results(task)


class MultiprocessingWorker(Worker):
    """Class of basic multiprocessing worker."""
    def start_loop(self, jobs: JoinableQueue, results: Queue):
        """
        Loop starting will be called by multiprocessing solver.

        :param JoinableQueue jobs: Queue for jobs
        :param Queue results: Queue for results
        """
        logger.info('%s %s Starting',
                    multiprocessing.current_process().name,
                    socket.gethostname())
        self.start()
        self._loop(jobs, results)

    def _loop(self, jobs: JoinableQueue, results: Queue):
        logger.info('%s %s Entering the loop',
                    multiprocessing.current_process().name,
                    socket.gethostname())
        while True:
            (i, task) = jobs.get()
            logger.debug('%s %s Starting doing the job %i',
                         multiprocessing.current_process().name,
                         socket.gethostname(),
                         i)
            try:
                res = self.do_the_job(task)
                results.put((i, res))
                jobs.task_done()
            except Exception as err:
                results.put((-1, err))
                raise err


class MPIWorker(Worker):
    """Class of basic MPI worker"""
    def start_loop(self, comm):
        """Loop start will be called by MPI solver.

        :param comm: Intracomm, COMM_WORLD commonly
        """
        import mpi4py
        if not isinstance(comm, mpi4py.MPI.Intracomm):
            raise TypeError('comm has to be instance of mpi4py.MPI.Intracomm')
        logger.info('Rang %i %s Starting',
                    comm.Get_rank(),
                    socket.gethostname())
        self.start()
        self._loop(comm)

    def _loop(self, comm):
        logger.info('Rang %i %s Entering the loop',
                    comm.Get_rank(),
                    socket.gethostname())
        while True:
            req = comm.irecv(source=0)
            (i, task) = req.wait()
            if i is None and task is None:
                return
            logger.debug('Rang %i %s Starting doing the job %i',
                         comm.Get_rank(),
                         socket.gethostname(),
                         i)
            res = self.do_the_job(task)
            req = comm.isend((i, res), dest=0)
            req.wait()



"""
This module contains cache classes
"""
from abc import ABCMeta, abstractmethod
from typing import Generic, Dict
from ._typevars import Task, Result


class BaseCache(Generic[Task, Result], metaclass=ABCMeta):
    @abstractmethod
    def __contains__(self, item: Task) -> bool:
        """
        Returns `True` if hashable item is in the cache, `False` otherwise.

        :param item:
        :return:
        """

    @abstractmethod
    def __getitem__(self, item: Task) -> Result:
        """
        Returns result by hashable item.
        If result is not in the cache raises `KeyError`.

        :param item:
        :return:
        """

    @abstractmethod
    def __setitem__(self, item: Task, value: Result):
        """
        Set result by hashable item.

        :param item:
        :param value:
        :return:
        """

    def get(self, item: Task) -> Result:
        """
        Returns result by hashable item.
        If result is not in the cache returns `None`.

        :param item:
        :return:
        """
        if item in self:
            return self[item]
        return None


class DictCache(BaseCache):
    """Simple cache based on dict."""
    def __init__(self):
        self._cache: Dict[Task, Result] = dict()

    def __getitem__(self, item: Task) -> Result:
        return self._cache.get(item)

    def __setitem__(self, item: Task, value: Result):
        self._cache[item] = value

    def __contains__(self, item: Task) -> bool:
        return item in self._cache




from pathlib import Path
import mph
from .model import ComsolModel

from ..workers import Worker, MultiprocessingWorker, MPIWorker
from .._typevars import Task, Result

import logging

logger = logging.getLogger(__package__ + '.worker')


# noinspection PyMissingOrEmptyDocstring
class ComsolWorker(Worker):
    def __init__(
            self,
            model: ComsolModel,
            filepath: str or Path,
            mph_options: dict = None,
            client_args=None,
            client_kwargs=None
    ):
        if not isinstance(model, ComsolModel):
            raise TypeError('Model has to be an object of ComsolModel class')

        super(ComsolWorker, self).__init__(model=model)

        self.client = None
        self.model = model

        self._mph_options = {} if mph_options is None else mph_options
        self._client_args = [] if client_args is None else client_args
        self._client_kwargs = {} if client_kwargs is None else client_kwargs
        self._filepath = filepath

    def start(self):
        for option in self._mph_options:
            mph.option(option, self._mph_options[option])
        logger.debug(f'Opening COMSOL model {self._filepath}')
        self.client = mph.start(*self._client_args, **self._client_kwargs)  # type: mph.client
        self.model.java = self.client.load(self._filepath).java
        self.model.configure()

    def do_the_job(self, task: Task) -> Result:
        self.model.pre_build(task)
        self.model.build()
        self.model.pre_solve(task)
        self.model.mesh()
        self.model.solve()
        results = self.model.results(task)
        self.model.pre_clear(task)
        self.model.clear()
        return results


# noinspection PyMissingOrEmptyDocstring
class ComsolMultiprocessingWorker(ComsolWorker, MultiprocessingWorker):
    def start(self):
        super().start()


# noinspection PyMissingOrEmptyDocstring
class ComsolMPIWorker(ComsolWorker, MPIWorker):
    def start(self):
        super().start()
